import numpy as np
import pandas as pd
from sklearn.decomposition import PCA
from sklearn.preprocessing import StandardScaler

from jupyter_analysis_extension.statistics_operation.common_util import CommonUtil
from jupyter_analysis_extension.statistics_operation.statistic_original import StatisticOperation
from jupyter_analysis_extension.statistics_widget.basic.basic_pca import WidgetBasicPCA
from jupyter_analysis_extension.statistics_widget.basic.basic_reliability import WidgetBasicReliability
from jupyter_analysis_extension.statistics_widget.common_widget import CommonWidget
from jupyter_analysis_extension.utils.widget_maker import WidgetMaker


class BasicPCA(StatisticOperation):
    def __init__(self, df, column_type_list, refresh_tab, logger):
        self.op_name = "Principal Component Analysis"
        self.sub = [
            {"name": "Variable(s)", "type": ["N"], "max_count": 0, "min_count": 1},
        ]
        self.option = [
            {"name": "Components Count", "type": "numberText", "value": 2},
            {"name": "Chart types", "type": "checkbox", "value": ['PCA plot','Scree plot', 'Matrix plot']},
            {"name": "Factor Analysis", "type": "checkbox", "value": ['KMO test', 'Bartlett test']},
        ]
        self.callback = {"refresh": refresh_tab}
        super(BasicPCA, self).__init__(self.op_name, df, column_type_list, refresh_tab, logger)

    def construct_control_panel(self):
        self.control_panel, self.control_data = CommonWidget.make_control_panel(
            df=self.input_df,
            type_list=self.column_type_list,
            title=self.op_name,
            sub=self.sub,
            option=self.option,
            callback=self.callback
        )

        self.column_value = self.control_data["column"]

        self.connect_button_layout(self.control_panel.children[1], self.on_analyze_button_clicked)

    def on_analyze_button_clicked(self, widget):
        self.set_analyze_text('In progress...')

        # catch invalid data
        is_valid = self.validate_input_data(self.control_data)
        if not is_valid:
            return

        # get analysis data & option

        var_layout_var_list = list(self.control_data['Variable(s)'])
        var_var_list = []
        for layout_var in var_layout_var_list:
            var_var_list.append(self.column_value[layout_var])

        option = self.control_data['option']
        option_components_count = option['Components Count']
        option_chart_types = option['Chart types']
        option_factor_analysis = option['Factor Analysis']

        result_widgets = self.get_result_widgets(var_var_list, option_components_count, option_chart_types,option_factor_analysis)

        CommonUtil.update_result_layout(self.control_panel, result_widgets)

        self.set_analyze_text('Done.')

    def validate_input_data(self, control_data):
        error_msg = CommonUtil.check_option_checkbox(control_data, 'Chart types', 0)
        if error_msg is not None:
            self.set_analyze_text(error_msg, 'warning')
            return False

        error_msg = CommonUtil.check_variables_shapes(control_data, 'Variable(s)', ['Number'],
                                                      self.column_type_list, self.column_value, self.sub)
        if error_msg is not None:
            self.set_analyze_text(error_msg, 'warning')
            return False

        if control_data['option']['Components Count'] > len(control_data['Variable(s)']) is not None:
            self.set_analyze_text('Components Count cannot be bigger than number of Variable(s)', 'warning')
            return False

        return True

    def get_result_widgets(self, var_var_list, option_components_count, option_chart_types,option_factor_analysis):

        X_ = StandardScaler().fit_transform(self.input_df[var_var_list])
        pca = PCA(n_components=len(var_var_list))
        components = pca.fit_transform(X_)
        explained_variance_ratio = pca.explained_variance_ratio_
        cum_explained_variance_ratio = np.cumsum(explained_variance_ratio)

        # Statistical Chart
        statistic_table_header = ['']
        statistic_table_body_item = ['Explained Variance(%)']
        statistic_table_body_item_2 = ['Cumulative EV(%)']
        for i in range(0, len(var_var_list)):
            statistic_table_header.append("PC" + str(i + 1))
            statistic_table_body_item.append(str(round(explained_variance_ratio[i] * 100, 2)) + "%")
            statistic_table_body_item_2.append(str(round(cum_explained_variance_ratio[i] * 100, 2)) + "%")

        statistic_table_body = [statistic_table_body_item, statistic_table_body_item_2]
        statistic_table_footer = "* Components Count " + str(option_components_count) + " is " + str(
            round(cum_explained_variance_ratio[option_components_count - 1] * 100, 2)) + "%"

        statistic_table = WidgetMaker.get_styled_table(statistic_table_header, statistic_table_body,
                                                       statistic_table_footer)

        # EDA Chart
        chart_accordion_title = "Exploratory Data Analysis(EDA) Chart"
        chart_list = []
        if option_chart_types['PCA plot']:
            chart_list.append(WidgetBasicPCA.get_pca_variance_chart(pca.explained_variance_ratio_))

        if option_chart_types['Scree plot']:
            chart_list.append(
                WidgetBasicPCA.get_scree_plot(pca.explained_variance_))

        if option_chart_types['Matrix plot']:
            pca_with_input = PCA(n_components=option_components_count)
            principal_components = pca_with_input.fit_transform(X_)
            columns_list = []
            for i in range(0, option_components_count):
                columns_list.append("PC" + str(i + 1))
            transform_df = pd.DataFrame(data=principal_components, columns=columns_list)
            chart_list.append(
                WidgetBasicPCA.get_matrix_plot(transform_df, columns_list, pca_with_input.explained_variance_ratio_))
        chart_accordion_item = WidgetMaker.get_vertical_box(box_items=chart_list,
                                                            style=dict(display="flex", width="100%",
                                                                       align_items="center", justify_content="space-around"))

        #  Factor Analysis(fa)
        fa_accordion_title = "Factor Analysis"
        fa_list = []
        if option_factor_analysis['KMO test']:
            kmo = self.calculate_kmo(self.input_df[var_var_list])
            kmo_table_header = ['', 'MSA']
            kmo_table_body = [['Overall',kmo[1]]]
            for index, item in enumerate(var_var_list):
                kmo_table_body.append([item, round(kmo[0][index],3)])

            kmo_table = WidgetMaker.get_styled_table(kmo_table_header, kmo_table_body,direation="row")
            standard_table = WidgetBasicPCA.get_kmo_standard_table()
            table_children = list(kmo_table.children)
            table_children.append(standard_table)
            kmo_table.children = tuple(table_children)
            kmo_table.layout=dict(display="flex", width="100%",align_items="center",flex_direction="row" ,justify_content="space-around")

        # if option_factor_analysis['Bartlett test']:
        #     fa_list.append(
        #         WidgetBasicPCA.get_scree_plot(pca.explained_variance_))
        #
        #
        # fa_accordion_item = WidgetMaker.get_vertical_box(box_items=fa_list,
        #                                                     style=dict(display="flex", width="100%",
        #                                                                align_items="center", justify_content="space-around"))

        return CommonWidget.get_accordion_widget([statistic_table, chart_accordion_item, kmo_table],
                                                 ['Statistical table', chart_accordion_title,fa_accordion_title])


    def calculate_kmo(self, x):
        """
        Calculate the Kaiser-Meyer-Olkin criterion
        for items and overall. This statistic represents
        the degree to which each observed variable is
        predicted, without error, by the other variables
        in the dataset. In general, a KMO < 0.6 is considered
        inadequate.
        Parameters
        ----------
        x : array-like
            The array from which to calculate KMOs.
        Returns
        -------
        kmo_per_variable : numpy array
            The KMO score per item.
        kmo_total : float
            The KMO score overall.
        """
        # calculate the partial correlations
        partial_corr = self.partial_correlations(x)

        # calcualte the pair-wise correlations
        x_corr = self.corr(x)

        # fill matrix diagonals with zeros
        # and square all elements
        np.fill_diagonal(x_corr, 0)
        np.fill_diagonal(partial_corr, 0)

        partial_corr = partial_corr**2
        x_corr = x_corr**2

        # calculate KMO per item
        partial_corr_sum = np.sum(partial_corr, axis=0)
        corr_sum = np.sum(x_corr, axis=0)
        kmo_per_item = corr_sum / (corr_sum + partial_corr_sum)

        # calculate KMO overall
        corr_sum_total = np.sum(x_corr)
        partial_corr_sum_total = np.sum(partial_corr)
        kmo_total = corr_sum_total / (corr_sum_total + partial_corr_sum_total)
        return kmo_per_item, kmo_total

    def partial_correlations(self, x):
        numrows, numcols = x.shape
        x_cov = self.cov(x, ddof=1)
        empty_array = np.empty((numcols, numcols))
        empty_array[:] = np.nan
        if numcols > numrows:
            icvx = empty_array
        else:
            try:
                assert np.linalg.det(x_cov) > np.finfo(np.float32).eps
                icvx = np.linalg.inv(x_cov)
            except AssertionError:
                icvx = np.linalg.pinv(x_cov)
            except np.linalg.LinAlgError:
                icvx = empty_array

        pcor = -1 * self.covariance_to_correlation(icvx)
        np.fill_diagonal(pcor, 1.0)
        return pcor

    def cov(self, x, ddof=0):
        r = np.cov(x, rowvar=False, ddof=ddof)
        return r


    def corr(self, x):
        x = (x - x.mean(0)) / x.std(0)
        r = self.cov(x)
        return r

    def covariance_to_correlation(self, m):
        numrows, numcols = m.shape
        if not numrows == numcols:
            raise ValueError('Input matrix must be square')

        Is = np.sqrt(1 / np.diag(m))
        retval = Is * m * np.repeat(Is, numrows).reshape(numrows, numrows)
        np.fill_diagonal(retval, 1.0)
        return retval



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) TmaxBI.
# Distributed under the terms of the Modified BSD License.
import ipywidgets as widgets

# Graph
import plotly.graph_objs as go


class WidgetANOVAANCOVA:
    @ staticmethod
    def get_statistic_result_table(table, degree_freedom, test_type):
        res_table_info = ''
        target_degree_freedom = degree_freedom


        '''
        statistic_value = round(item.statistic, 3)
        if statistic_tests_options[pairindex] == 'Welch\'s' or statistic_tests_options[pairindex] == 'Fisher\'s':
            target_degree_freedom = '-'

        if statistic_tests_options[pairindex] == 'Fisher\'s':
            statistic_value = '{0:e}'.format(item.statistic)

        res_table_info += """
                    <tr>
                        <td>{0}</td>
                        <td>{1}</td>
                        <td>{2}</td>
                        <td>{3}</td>
                        <td>{4}</td>
                    </tr>
                    """.format(name[pairindex], statistic_tests_options[pairindex], statistic_value,
                               target_degree_freedom, round(item.pvalue, 3))

        res_table = widgets.HTML(
            value="""
                <table style="width:100%">
                <tr style="border-bottom: 1px solid black;">
                    <th></th>
                    <th></th>
                    <th>Statistic</th>
                    <th>df</th>
                    <th>p-value</th>
                </tr>
               {0}
                </table>""".format(res_table_info),
            disabled=True
        )
        '''
        res_table = widgets.HTML(table.to_html())
        statistic_result_table = widgets.VBox([res_table])
        statistic_result_table.add_class('styledTable')

        return statistic_result_table

    @staticmethod
    def get_boxplot_sns(grouped_data, group_val_distinct):
        fig = go.FigureWidget()
        for data, index in zip(grouped_data, group_val_distinct):
            fig.add_trace(go.Box(y=data, name=index))

        fig.update_layout(
            width=500,
            height=500,
        )
        return fig


#!/usr/bin/env python
# coding: utf-8

# Graph
from itertools import chain, combinations

# Copyright (c) kyoungjun.
# Distributed under the terms of the Modified BSD License.
import ipywidgets as widgets
from ipyevents import Event
from ipywidgets.widgets.widget_layout import Layout

from jupyter_analysis_extension.statistics_operation.common_util import CommonUtil
from jupyter_analysis_extension.utils.icon_info import BASIC_ICON, T_TEST_ICON, ANOVA_ICON, REGRESSION_ICON, FREQUENCIES_ICON
from jupyter_analysis_extension.utils.widget_maker import WidgetMaker


class CommonWidget:
    @staticmethod
    def make_control_panel(df=None, type_list=None, title="Unknown", sub=None, option=None, callback=None):
        if callback is None:
            callback = {}
        if option is None:
            option = []
        if sub is None:
            sub = []
        if type_list is None:
            type_list = []

        control_data = {}
        # header
        header_widget = CommonWidget.get_header_widget(title)
        control_data["title"] = title

        column_name_type_map = CommonUtil.get_column_name_with_type(df.columns, type_list)
        control_data["column"] = column_name_type_map

        main_layout = CommonWidget.get_multiple(list(column_name_type_map), '300px')
        control_data["main"] = list(column_name_type_map)

        sub_item_list = []
        item_multiple_dict = {}
        add_btn_list = []
        delete_btn_list = []
        for item in sub:
            item_multiple = CommonWidget.get_multiple([], '100%')
            add_button = CommonWidget.get_add_button()
            delete_button = CommonWidget.get_delete_button()
            if item['max_count'] > 0:
                item_layout = CommonWidget.get_limit_multiple_with_button_layout(item["name"], item_multiple,
                                                                                 item['max_count'], add_button,
                                                                                 delete_button)
            else:
                item_layout = CommonWidget.get_multiple_with_button_layout(item["name"], item_multiple, add_button,
                                                                           delete_button)
            item_multiple_dict[item["name"]] = item_multiple
            add_btn_list.append(add_button)
            delete_button.layout = Layout(width="40px", display="none")
            delete_btn_list.append(delete_button)
            CommonUtil.connect_multiples_by_button(
                control_data, "main", main_layout, item["name"], item_multiple, add_button, delete_button,
                max_count=item['max_count'])
            sub_item_list.append(item_layout)

        def on_main_selected(value):
            for btn in delete_btn_list:
                btn.layout = Layout(width="40px", display="none", margin="0px")
            for btn in add_btn_list:
                btn.layout = Layout(width="40px", display="",margin="0px")

        main_layout.observe(on_main_selected, names='value')

        def on_sub_selected(value):
            for btn in add_btn_list:
                btn.layout = Layout(width="40px", display="none",margin="0px")
            for btn in delete_btn_list:
                btn.layout = Layout(width="40px", display="",margin="0px")

        for key in item_multiple_dict.keys():
            item_multiple_dict[key].observe(on_sub_selected, names='value')

        # combine sub
        sub_layout = CommonWidget.get_sub_layout(sub_item_list, "300px")

        # layout
        column_layout = CommonWidget.get_column_layout(
            header_widget, main_layout, sub_layout)

        # option layout
        option_left_item_list = []
        option_center_item_list = []
        option_right_item_list = []
        control_data["option"] = {}
        for item in option:
            if item["type"] == "checkbox":
                item_widget = CommonWidget.get_checkbox(control_data["option"], item["name"], item["value"])

            elif item["type"] == "radio":
                item_widget = CommonWidget.get_radio_buttons(control_data["option"], item["name"], item["value"], 0)

            elif item["type"] == "numberText":
                item_widget = CommonWidget.get_number_text(control_data["option"], item["name"], item["value"])

            elif item["type"] == "numberText-radio":
                widget_value = item["value"]
                item_widget = CommonWidget.get_number_field_and_radio_buttons(
                    control_data["option"],
                    item["name"],
                    widget_value["numberText"]["name"],
                    widget_value["numberText"]["value"],
                    widget_value["radio"]["value"], 0)

            elif item["type"] == "connected_multiple":
                connect_target_widgets = []
                for target in item["connect"]:
                    connect_target_widgets.append(item_multiple_dict[target])
                item_widget = CommonWidget.get_connected_multiple(control_data["option"], item["name"],
                                                                         item["acc_title"],
                                                                         item["value"], connect_target_widgets,
                                                                         column_name_type_map, type_list)

            elif item["type"] == "connected_unique_dropdown":
                connect_target_widget = item_multiple_dict[item["connect"]]
                item_widget = CommonWidget.get_connected_dropdown(control_data["option"], item["name"],
                                                                         connect_target_widget, column_name_type_map,
                                                                         type_list, df=df,
                                                                         place_holder=str(
                                                                             "Check the value of " + item["connect"]))

            elif item["type"] == "individual_connected_unique_dropdown":
                connect_target_widgets = []
                for connect_item in item["connect"]:
                    connect_target_widgets.append(item_multiple_dict[connect_item])

                acc_title = item["acc_title"] if "acc_title" in item else item["name"]
                item_widget = CommonWidget.get_individual_connected_dropdown(control_data["option"],
                                                                                               item["name"],
                                                                                               connect_target_widgets,
                                                                                               column_name_type_map,
                                                                                               type_list, df=df,
                                                                                               acc_title=acc_title)

            elif item["type"] == "individual_connected_unique_numberField":
                connect_target_widgets = []
                for connect_item in item["connect"]:
                    connect_target_widgets.append(item_multiple_dict[connect_item])

                acc_title = item["acc_title"] if "acc_title" in item else item["name"]
                item_widget = WidgetMaker.get_individual_connected_unique_numberField(control_data["option"],
                                                                                               item["name"],
                                                                                               connect_target_widgets,
                                                                                               column_name_type_map,
                                                                                               type_list, df=df,
                                                                                               acc_title=acc_title)
            else:
                print("Type is incorrect in option widgets.")
                item_widget = None

            if item_widget is not None:
                widget_with_label = CommonWidget.get_widget_with_label(item["name"], item_widget, margin="0px 0px 0px 5px")

                if "position-column" in item:
                    if item["position-column"] == "center":
                        option_center_item_list.append(widget_with_label)
                    elif item["position-column"] == "right" :
                        option_right_item_list.append(widget_with_label)
                    else:
                        option_left_item_list.append(widget_with_label)
                else:
                    option_left_item_list.append(widget_with_label)

        option_layout = CommonWidget.get_option_layout(option_left_item_list,option_center_item_list,option_right_item_list )

        # analyze layout
        analyze_button = CommonWidget.get_button_with_icon('analyze')
        refresh_button = CommonWidget.get_button_with_icon('refresh')

        def on_refresh_button_clicked(self):
            if "refresh" in callback:
                callback["refresh"](title)

        refresh_button.on_click(on_refresh_button_clicked)

        analyze_textfield = CommonWidget.get_textfield()
        refresh_textfield = CommonWidget.get_textfield()

        analyze_layout = CommonWidget.get_analyze_layout([analyze_textfield, refresh_textfield],
                                                         [analyze_button, refresh_button])

        result_layout = CommonWidget.get_result_layout()

        control_panel = CommonWidget.get_control_panel_widget(
            column_layout,
            option_layout,
            analyze_layout,
            result_layout)
        return control_panel, control_data

    @staticmethod
    def get_dropdown_with_description(title, options, default_option):
        if default_option not in options:
            default_option = options[0]

        return widgets.Dropdown(
            options=options,
            value=default_option,
            description=title,
            layout=widgets.Layout(height="99%", width="300px",margin="0"),
            style=dict(description_width='150px')
        )

    @staticmethod
    def get_dropdown(options, default_option):
        if default_option not in options:
            default_option = options[0]

        return widgets.Dropdown(
            options=options,
            value=default_option,
            layout=widgets.Layout(height="99%", width="200px",margin="0"),
            style=dict(description_width='100px')
        )

    #
    # layout widget
    #
    # multiple widget #
    @staticmethod
    def get_multiple(columns, height, width="95%"):
        return widgets.SelectMultiple(
            options=columns,
            disabled=False,
            style=dict(description_width='initial'),
            layout=widgets.Layout(
                height=height, width=width, margin='0')
        )

    @staticmethod
    def get_limit_multiple_with_button_layout(label, target_multiple, max_count, add_button, delete_button):
        layout_label = widgets.HTML(
            value=label, layout=widgets.Layout(height="25px", margin="0px"))
        layout_label.add_class("multiple-select-title")
        hidden_button = CommonWidget.get_hidden_button()

        max_height = max_count * 20 + 33
        multiple_height = max_count * 20 + 8
        target_multiple.layout = Layout(width="95%", height=str(multiple_height) + "px", margin="0")
        return widgets.VBox([
            widgets.HBox([hidden_button, layout_label],
                         layout=widgets.Layout(display="flex", justify_content="flex-start", min_height="25px",
                                               margin="0")
                         ),
            widgets.HBox(
                [widgets.VBox([add_button, delete_button],
                              layout=widgets.Layout(display="flex", justify_content="flex-start", width="45px",
                                                    height="100%", margin="0")),
                 widgets.VBox([target_multiple],
                              layout=widgets.Layout(display="flex", align_items="flex-start", width="calc(100% - 45px)",
                                                    height="100%", margin="0"))
                 ],
                layout=widgets.Layout(height="100%", display="flex", margin="0")
            )
        ], layout=widgets.Layout(min_height=str(max_height) + "px", width="300px", display="flex"))

    @staticmethod
    def get_multiple_with_button_layout(label, target_multiple, add_button, delete_button, width="300px"):
        layout_label = widgets.HTML(
            value=label, layout=widgets.Layout(height="25px", margin="0px"))
        layout_label.add_class("multiple-select-title")
        hidden_button = CommonWidget.get_hidden_button()

        return widgets.VBox([
            widgets.HBox([hidden_button, layout_label],
                         layout=widgets.Layout(display="flex", justify_content="flex-start", min_height="25px",
                                               margin="0")
                         ),
            widgets.HBox(
                [widgets.VBox([add_button, delete_button],
                              layout=widgets.Layout(display="flex", justify_content="flex-start", width="45px",
                                                    height="100%",margin="0")),
                 widgets.VBox([target_multiple],
                              layout=widgets.Layout(display="flex", align_items="flex-start", width="calc(100% - 45px)",
                                                    height="100%", margin="0"))
                 ],
                layout=widgets.Layout(height="100%",display="flex", margin="0")
            )
        ], layout=widgets.Layout(height="100%", width=width, display="flex", margin="0", padding="1px 0px 1px 0px"))

    @staticmethod
    def get_multiple_with_label(label, target_multiple, width="300px"):
        layout_label = widgets.HTML(
            value=label, layout=widgets.Layout(height="25px", margin="0"))
        layout_label.add_class("multiple-select-title")

        return widgets.HBox([
            widgets.VBox([layout_label, target_multiple],
                         layout=widgets.Layout(display="flex", align_items="flex-start", width="100%", height="100%",
                                               margin="0"))
        ], layout=widgets.Layout(height="100%", width=width, display="flex", margin="0"))

    # button widget #
    @staticmethod
    def get_add_button():
        return widgets.Button(
            disabled=False,
            tooltip='add',
            icon='arrow-right',
            layout=widgets.Layout(width="40px", margin="0px"))

    @staticmethod
    def get_hidden_button():
        return widgets.Button(
            disabled=False,
            tooltip='add',
            icon='refresh',
            layout=widgets.Layout(width="45px", visibility="hidden", height="25px", margin="0px"))

    @staticmethod
    def get_delete_button():
        return widgets.Button(
            disabled=False,
            button_style='',
            tooltip='delete',
            icon='arrow-left',
            layout=widgets.Layout(width="40px", margin="0px"))

    @staticmethod
    def get_button_with_icon(icon_type):
        if icon_type == 'analyze':
            button_widget = widgets.Button(
                description='Analyze',
                disabled=False,
                button_style='',
                tooltip='Analyze',
                icon='search',
            )

        elif icon_type == 'refresh':
            button_widget = widgets.Button(
                description='Refresh',
                disabled=False,
                button_style='',
                tooltip='Refresh',
                icon='refresh',
            )
        elif icon_type == 'export-html':
            button_widget = widgets.Button(
                description='Export to HTML',
                disabled=False,
                button_style='',
                tooltip='Export to HTML',
                icon='browser',
            )
        elif icon_type == 'export-pdf':
            button_widget = widgets.Button(
                description='Export to PDF',
                disabled=False,
                button_style='',
                tooltip='Export to PDF',
                icon='file-pdf',
            )
        else:
            button_widget = None
            exit(1)
        button_widget.add_class("round-button-5")
        return button_widget

    @staticmethod
    def get_warning_textfield(warning_text):
        return widgets.HTML(
            value=warning_text)

    @staticmethod
    def get_textfield(text="&nbsp"):
        return widgets.HTML(
            value="""<div style="display:flex;justify-content:flex-end">{0}</div>""".format(
                text))

    @staticmethod
    def get_hbox(items, justify_content="center"):
        return widgets.HBox(
            items,
            layout=widgets.Layout(display="flex", width="100%", min_width="600px", justify_content=justify_content),
        )

    @staticmethod
    def get_vbox(items, justify_content="center"):
        return widgets.VBox(
            items,
            layout=widgets.Layout(display="flex", width="100%", min_width="600px", justify_content=justify_content),
        )

    @staticmethod
    def get_header_widget(test_name):
        children = [
            widgets.HTML(
                value="""<div style="font-size:2rem">{0}</div>""".format(test_name)),
        ]
        box_layout = Layout(display='flex',
                            border_bottom='1px solid gray',
                            width='100%')
        return widgets.Box(children=children, layout=box_layout)

    @staticmethod
    def get_analyze_layout(text_field, button):
        text_list = []
        button_list = []
        if isinstance(text_field, list):
            text_list = text_field
        else:
            text_list.append(text_field)

        if isinstance(button, list):
            button_list = button
        else:
            button_list.append(button)

        return widgets.HBox(
            [widgets.VBox(text_list), widgets.VBox(button_list)],
            layout=widgets.Layout(display="flex", height="100%", align_items="flex-end", width="100%",
                                  justify_content="flex-end"))

    @staticmethod
    def get_sub_layout(target, height="100%"):
        return widgets.VBox(target,
                            layout=widgets.Layout(display="flex", height=height, align_items="flex-end", width="100%"))

    @staticmethod
    def get_column_layout(header, var_layout, target_layout):
        return widgets.VBox([
            header,
            widgets.HBox(
                [var_layout, target_layout],
                layout=widgets.Layout(display="flex", width="100%", min_width="600px", justify_content="flex-end")),
        ])

    @staticmethod
    def get_option_layout(option_left_item_list,option_center_item_list,option_right_item_list):
        children = [
            widgets.HTML(
                value="""<div style="font-size:2rem; padding-top:calc(2rem + 4px)">Option</div>"""),
        ]
        option_title = [widgets.Box(children=children, layout=widgets.Layout(display='flex',
                            border_bottom='1px solid gray',
                            width='100%'))]
        layout_children = []
        if len(option_left_item_list) + len(option_center_item_list) +  len(option_right_item_list)> 0:
            left_list = widgets.VBox(option_left_item_list, layout=widgets.Layout(display="flex"))
            center_list = widgets.VBox(option_center_item_list, layout=widgets.Layout(display="flex"))
            right_list = widgets.VBox(option_right_item_list, layout=widgets.Layout(display="flex"))
            option_layout = widgets.HBox([left_list, center_list, right_list])
            layout_children = option_title + [option_layout]
        return widgets.VBox(layout_children,
                            layout=widgets.Layout(display="flex", height="100%", min_height="300px",
                                                  justify_content="flex-start"))

    @staticmethod
    def get_result_layout():
        return widgets.VBox([], layout=widgets.Layout(display='flex', justify_content='center', width="100%"))

    @staticmethod
    def get_control_panel_widget(colmun_layout, option_layout, analyze_layout, result_layout):
        default_style = Layout(display='flex', width="99%",
                               align_items='flex-start', justify_content="center")
        control_panel_widget = widgets.VBox(
            [widgets.HBox([colmun_layout, option_layout], layout=default_style), analyze_layout, result_layout],
            layout=default_style)

        return control_panel_widget

    # @staticmethod
    # def get_box_plot_plotly(data_list):
    #     f = go.FigureWidget()
    #     for data in data_list:
    #         f.add_trace(go.Box(y=data, name=data.name))
    #     return f

    @staticmethod
    def get_accordion_widget(widget_list, title_list, selected_index=0, min_width="10px"):
        accordion_widget = widgets.Accordion(
            children=widget_list,
            layout=widgets.Layout(width="100%", display="flex", justify_content='center', margin="0", min_width=min_width),
            selected_index=selected_index)
        for i in range(len(accordion_widget.children)):
            accordion_widget.set_title(i, title_list[i])
        return accordion_widget

    @staticmethod
    def get_widget_with_label(label_text, widget, orientation="v", margin="0"):
        label = widgets.HTML(
            value="""<div style="font-size:1.5rem; font-weight:bold">{0}</div>""".format(label_text),
            layout=widgets.Layout(width="100%", min_height="25px", margin="0"))
        if orientation == "v":
            result = widgets.VBox(
                [label, widget],
                layout=widgets.Layout(margin=margin))
        else:
            result = widgets.HBox(
                [label, widget],
                layout=widgets.Layout(margin=margin))
        return result


    @staticmethod
    def get_file_meta_widget(file_name, file_size, file_type):
        format_size = str(round(file_size/1024)) + "KB" if file_size < 1024*1024 else str(round(file_size/1024/1024,2)) + "MB"
        file_name_widget = widgets.HTML(
            value=file_name,
            description='Name : ',
            style=dict(description_width='70px')
        )
        file_size = widgets.HTML(
            value=format_size,
            description='Size : ',
            style=dict(description_width='70px')
        )
        file_type = widgets.HTML(
            value=file_type,
            description='Type : ',
            style=dict(description_width='70px')
        )
        file_meta_widget = widgets.VBox([file_name_widget,file_size,file_type], layout=widgets.Layout(margin='0px 0px 0px 10px'))
        return file_meta_widget
    #
    # option widget
    #
    @staticmethod
    def get_radio_buttons(control_data_option, title, values, default_value_index=0):
        control_data_option[title] = {}

        def radio_changed(radio_value):
            control_data_option[title]["radio"] = radio_value['new']

        radio_buttons = widgets.RadioButtons(
            options=values,
            value=values[default_value_index],
            disabled=False,
            layout=widgets.Layout(margin='0px 0px 0px 5px')
        )
        radio_buttons.observe(radio_changed, names="value")
        control_data_option[title]["radio"] = values[default_value_index]
        return widgets.VBox([radio_buttons], layout=widgets.Layout(margin='0px 0px 0px 10px'))    \

    @staticmethod
    def get_number_text(control_data_option, title, default_value=0):
        control_data_option[title] = {}

        def number_changed(value):
            control_data_option[title]= value['new']


        widget = widgets.BoundedIntText(
            value=default_value,
            min=0,
            step=1,
            disabled=False,
            layout=widgets.Layout(margin='0px 0px 0px 5px', width="100px")
        )
        widget.observe(number_changed, names="value")
        control_data_option[title] = default_value
        return widgets.VBox([widget], layout=widgets.Layout(margin='0px 0px 0px 10px'))

    @staticmethod
    def get_checkbox(control_data_option, title, values):
        control_data_option[title] = {}

        def checkbox_clicked(checkbox_value):
            control_data_option[title][checkbox_value['owner'].description] = checkbox_value['new']

        checkbox_list = []
        for value in values:
            checkbox_widget = widgets.Checkbox(
                value=False,
                description=value,
                disabled=False,
                indent=False,
            )
            control_data_option[title][value] = False
            checkbox_widget.observe(checkbox_clicked, names="value")
            checkbox_widget.add_class("styled-check-box")
            checkbox_list.append(checkbox_widget)
        return widgets.VBox(checkbox_list, layout=widgets.Layout(margin='0px 0px 0px 12px'))

    @staticmethod
    def get_popup_menu(op_category, items, click_event):
        icon_op_name = ""
        if op_category == "Basic":
            icon_src = "data:image/svg+xml;base64,{0}".format(BASIC_ICON)
            icon_op_name = "기본"
        elif op_category == "T-test":
            icon_src = "data:image/svg+xml;base64,{0}".format(T_TEST_ICON)
            icon_op_name = "T-검정"
        elif op_category == "ANOVA":
            icon_src = "data:image/svg+xml;base64,{0}".format(ANOVA_ICON)
            icon_op_name = "분산분석"
        elif op_category == "Regression":
            icon_src = "data:image/svg+xml;base64,{0}".format(REGRESSION_ICON)
            icon_op_name = "회귀분석"
        elif op_category == "Frequencies":
            icon_src = "data:image/svg+xml;base64,{0}".format(FREQUENCIES_ICON)
            icon_op_name = "빈도분석"
        else :
            icon_src = "data:image/svg+xml;base64,{0}".format(BASIC_ICON)

        item_widgets = []

        for item in items:
            item_widgets.append(widgets.HTML(
                value="""
                <a>{0}</>
                """.format(item),
            ))
        drop_btn = widgets.HTML(
            value="""
            <img class="drop-img" src="{0}" alt="statistic icon"/>
            <div class="icon-op-name">&nbsp{1}▼</div>
            """.format(icon_src, icon_op_name)
        )
        drop_btn.add_class("dropdown-btn")
        dropdown_content = CommonWidget.get_vbox(item_widgets)
        dropdown_content.layout = widgets.Layout()

        dropdown_content.add_class("dropdown-content")

        dropdown = CommonWidget.get_vbox([drop_btn, dropdown_content])
        dropdown.add_class("dropdown")
        dropdown.layout = widgets.Layout(margin="5px 20px 5px 5px")  # margin 줘서 기존 속성 날림

        def handle_event(index):
            return lambda event: click_event(items[index])

        for index, item_widget in enumerate(item_widgets):
            d = Event(source=item_widget, watched_events=['click'])
            d.on_dom_event(handle_event(index))

        return dropdown

    @staticmethod
    def get_number_field_and_radio_buttons(control_data_option, title, text_title, text_value, radio_values,
                                           default_radio_value_index):
        control_data_option[title] = {}
        control_data_option[title]["number"] = 0
        control_data_option[title]["radio"] = radio_values[default_radio_value_index]

        def textfield_changed(value):
            control_data_option[title]["number"] = value['new']

        def radio_changed(value):
            control_data_option[title]["radio"] = value['new']

        number_field = widgets.FloatText(
            value=text_value,
            description=text_title,
            disabled=False,
            layout=widgets.Layout(width="180px"),
            style=dict(description_width='min-content')
        )
        number_field.observe(textfield_changed, names="value")

        radio_buttons = widgets.RadioButtons(
            options=radio_values,
            value=radio_values[default_radio_value_index],
            disabled=False,
            layout=widgets.Layout(margin='0px 0px 0px 5px')
        )
        control_data_option[title + "-radio"] = radio_values[default_radio_value_index]

        radio_buttons.observe(radio_changed, names="value")
        return widgets.VBox([number_field, radio_buttons],
                            layout=widgets.Layout(margin='0px 0px 0px 10px'))

    @staticmethod
    def get_connected_multiple(control_data_option, title, accordion_title, multiple_titles, connect_widgets,
                               column_name_type_map, type_list):
        delete_btn_list = []
        add_btn_list = []

        main_multiple = CommonWidget.get_multiple([], height='180px',width="200px")
        main_layout = CommonWidget.get_multiple_with_label(multiple_titles[0], main_multiple, width="205px")
        sub_multiple = CommonWidget.get_multiple([],height='180px',width="200px")
        add_btn = CommonWidget.get_add_button()
        delete_btn = CommonWidget.get_delete_button()
        sub_layout = CommonWidget.get_multiple_with_button_layout(multiple_titles[1], sub_multiple,add_btn,delete_btn, width="250px" )
        add_btn_list.append(add_btn)
        delete_btn.layout = Layout(width="40px", display="none")
        delete_btn_list.append(delete_btn)
        CommonUtil.connect_multiples_by_button(
            control_data_option, multiple_titles[0], main_multiple, multiple_titles[1], sub_multiple, add_btn,
            delete_btn)

        def on_main_selected(value):
            for btn in delete_btn_list:
                btn.layout = Layout(width="40px", display="none", margin="0")
            for btn in add_btn_list:
                btn.layout = Layout(width="40px", display="", margin="0")

        main_multiple.observe(on_main_selected, names='value')

        def on_sub_selected(value):
            for btn in add_btn_list:
                btn.layout = Layout(width="40px", display="none", margin="0")
            for btn in delete_btn_list:
                btn.layout = Layout(width="40px", display="", margin="0")

        sub_multiple.observe(on_sub_selected, names='value')

        def on_multiple_changed(value):
            multiples_options = []
            distinct_options = []
            multiple_option = []
            for multiple in connect_widgets:
                multiples_options.extend(list(multiple.options))

            for option in list(set(multiples_options)):
                column_name = column_name_type_map[option]
                if type_list[column_name] != "Number":
                    distinct_options.append(column_name)
                else:
                    multiple_option.append(column_name)

            combi = chain.from_iterable(combinations(distinct_options, r) for r in range(1, len(distinct_options) + 1))
            for item in list(combi):
                multiple_option.append(" ※ ".join(item))
            main_multiple.options = tuple(multiple_option)
            sub_multiple.options = ()

        main_multiple.observe(on_main_selected, names='value')

        for widget in connect_widgets:
            widget.observe(on_multiple_changed, names="options")

        accordion_item = widgets.HBox(
            [main_layout, sub_layout],
            layout=widgets.Layout(display="flex", justify_content="flex-start"))
        connected_multiple = CommonWidget.get_accordion_widget([accordion_item], [accordion_title], selected_index=None, min_width="480px")
        return widgets.VBox([connected_multiple], layout=widgets.Layout(margin='0px 0px 0px 10px'))

    @staticmethod
    def get_connected_dropdown(control_data_option, title, connect_widget, column_name_type_map, type_list, df,
                               place_holder="empty"):
        dropdown_options = [place_holder]
        connected_dropdown = widgets.Dropdown(
            options=dropdown_options,
            value=dropdown_options[0],
            disabled=False,
            layout=widgets.Layout(width="200px")
        )

        def on_dropdown_changed(value):
            control_data_option[title] = connected_dropdown.value

        connected_dropdown.observe(on_dropdown_changed, names="value")

        def on_multiple_changed(value):
            if len(list(connect_widget.options)) > 0:
                for option in list(connect_widget.options):
                    column_name = column_name_type_map[option]
                    type_name = type_list[column_name]
                    if type_name in ['Object', '2-Category', 'N-Category']:
                        unique_items = df[column_name].dropna().unique()
                        connected_dropdown.options = unique_items
                        connected_dropdown.value = unique_items[0]
                    else:
                        connected_dropdown.options = ["Select categorical variable for dependent variable"]
                        connected_dropdown.value = "Select categorical variable for dependent variable"
            else:
                connected_dropdown.options = dropdown_options
                connected_dropdown.value = dropdown_options[0]
            control_data_option[title] = connected_dropdown.value

        connect_widget.observe(on_multiple_changed, names="options")

        return connected_dropdown

    @staticmethod
    def get_individual_connected_dropdown(control_data_option, title, connect_widgets, column_name_type_map, type_list,
                                          df, acc_title):
        individual_connected_dropdown = None
        control_data_option[title] = {}

        def on_dropdown_changed(value):
            control_data_option[title][value['owner'].description] = value['new']

        def on_multiple_changed(value):
            individual_connected_dropdown.children = []
            total_options = []
            for item in connect_widgets:
                total_options.extend(list(item.options))

            dropdown_widgets = []
            changed_option = {}
            for option in total_options:
                column_name = column_name_type_map[option]
                if type_list[column_name] not in ['2-Category', 'N-Category', 'Object']:
                    unique_items = ['Select categorical variable']
                else:
                    unique_items = df[column_name].dropna().unique()
                value = unique_items[0]
                if column_name in control_data_option[title]:
                    value = control_data_option[title][column_name]
                dropdown_widget = CommonWidget.get_dropdown_with_description(title=column_name, options=unique_items,
                                                                             default_option=value)

                changed_option[column_name] = value
                dropdown_widgets.append(dropdown_widget)
                dropdown_widget.observe(on_dropdown_changed, names="value")
                individual_connected_dropdown.children = dropdown_widgets

            control_data_option[title] = changed_option

        for widget in connect_widgets:
            widget.observe(on_multiple_changed, names="options")

        individual_connected_dropdown = widgets.VBox([])

        accordion_item = widgets.HBox(
            [individual_connected_dropdown],
            layout=widgets.Layout(display="flex", justify_content="flex-start", margin="0", width="calc(100% + 5px)"))
        connected_multiple = CommonWidget.get_accordion_widget([accordion_item], [acc_title], selected_index=None, min_width="350px")

        return widgets.VBox([connected_multiple], layout=widgets.Layout(margin='0px 0px 0px 10px'))

    @staticmethod
    def get_export_layout(*args):
        margin = widgets.Layout(margin='5px 5px 5px 5px', align_items='center')
        agg_widget = widgets.VBox([widgets.HBox(list(args), layout=margin)], layout=margin)

        return agg_widget





import numpy as np


def preprocess_values(df):
    # Preprocess the data for inf or nan values
    # Draft ver: only remove the rows
    # TODO.
    # Should be changed to user can choose the methods.

    df = df.replace([np.inf, -np.inf], np.nan)
    df = df.dropna()
    return df



import plotly.graph_objs as go
import scipy.stats as stats


class ChartMaker:
    @staticmethod
    def get_qq_plot_plotly(name, data):
        qqplot_data = stats.probplot(data, dist="norm")
        fig = go.FigureWidget()

        fig.add_trace({
            'type': 'scatter',
            'x': qqplot_data[0][0],
            'y': qqplot_data[0][1],
            'mode': 'markers',
        })

        # TO-Do 45도 기울기의 그래프 그리는거
        # fig.add_trace({
        #     'type': 'scatter',
        #     'x': [qqplot_data[0][0][0], qqplot_data[0][0][-1]],
        #     'y': qqplot_data[1][0:2],
        #     'mode': 'lines',
        # })

        fig.update_layout(
            title=name+" qq-plot",
            title_x=0.5,
            title_font_size=20,
            width=400,
            height=400,
            xaxis_title = "Theoritical Quantities",
            xaxis_zeroline = False,
            yaxis_title="Sample Quantities",
            showlegend = False
        )
        return fig

    @staticmethod
    def get_box_plot_plotly(data_list):
        fig = go.FigureWidget()
        for data in data_list:
            fig.add_trace(
                go.Box(y=data,
                       name=data.name,
                       boxpoints='all',
                       jitter=0.2,
                       pointpos=-1.8
                       ))

        chart_width = 300 + len(data_list)*50
        fig.update_layout(
            width=chart_width,
            height=500,
        )
        return fig



#!/usr/bin/env python
# coding: utf-8

# Copyright (c) kyoungjun.
# Distributed under the terms of the Modified BSD License.
import pandas as pd
from scipy import stats

from jupyter_analysis_extension.statistics_operation.common_util import CommonUtil
from jupyter_analysis_extension.statistics_operation.statistic_original import StatisticOperation
from jupyter_analysis_extension.statistics_widget.basic.basic_analysis import WidgetBasicAnalysis
from jupyter_analysis_extension.statistics_widget.common_widget import CommonWidget
from jupyter_analysis_extension.statistics_widget.frequencies.frequencies_chi_square import WidgetFrequenciesChiSquare


class BasicAnalysis(StatisticOperation):
    def __init__(self, df, column_type_list, refresh_tab, logger):
        self.op_name = "Basic : Analysis"
        self.sub = [
            {"name": "Variable(s)", "type": ["N"], "max_count": 0, "min_count": 1},
        ]
        self.option = [
            {"name": "Statistical table", "type": "checkbox", "value": ['Base(기초)']},
            {"name": "Chart Types", "type": "checkbox", "value": ['Box plot']}
        ]
        self.callback = {"refresh": refresh_tab}

        super(BasicAnalysis, self).__init__(self.op_name, df, column_type_list, refresh_tab, logger)

    def construct_control_panel(self):
        self.control_panel, self.control_data = CommonWidget.make_control_panel(
            df=self.input_df,
            type_list=self.column_type_list,
            title=self.op_name,
            sub=self.sub,
            option=self.option,
            callback=self.callback
        )

        self.column_value = self.control_data["column"]

        self.connect_button_layout(self.control_panel.children[1], self.on_analyze_button_clicked)

    def on_analyze_button_clicked(self, widget):
        self.set_analyze_text('In progress...')

        # catch invalid data
        is_valid = self.validate_input_data(self.control_data)
        if not is_valid:
            return

        # get analysis data & option

        var_layout_var_list = list(self.control_data['Variable(s)'])
        var_var_list = []
        for layout_var in var_layout_var_list:
            var_var_list.append(self.column_value[layout_var])

        option = self.control_data['option']
        option_statistucal_table = option['Statistical table']
        option_chart_types = option['Chart Types']

        result_widgets = self.get_result_widgets(var_var_list, [option_statistucal_table, option_chart_types])

        CommonUtil.update_result_layout(self.control_panel, result_widgets)

        self.set_analyze_text('Done.')

    def validate_input_data(self, control_data):

        error_msg = CommonUtil.check_option_checkbox(control_data, 'Statistical table', 1)
        if error_msg is not None:
            self.set_analyze_text(error_msg, 'warning')
            return False

        error_msg = CommonUtil.check_option_checkbox(control_data, 'Chart Types', 1)
        if error_msg is not None:
            self.set_analyze_text(error_msg, 'warning')
            return False

        error_msg = CommonUtil.check_variables_shapes(control_data, 'Variable(s)',
                                                      ['N-Category', '2-Category', 'Object', 'Number'],
                                                      self.column_type_list, self.column_value, self.sub)
        if error_msg is not None:
            self.set_analyze_text(error_msg, 'warning')
            return False

        return True

    def get_result_widgets(self, var_var_list, tests_options):
        first_col_values = []
        statistic_chart = []
        statistic_tests_options = []
        statistic_result = {}

        for var_var in var_var_list:
            statistic_chart.append(self.input_df[var_var])

        if tests_options[0]['Base(기초)']:
            statistic_result['Base(기초)'] = {}
            for rows_var in var_var_list:
                statistic_result['Base(기초)'][rows_var] = self.input_df[rows_var].describe()
                first_col_values.append(rows_var)

        df_count = len(self.input_df)

        statistic_result_table = WidgetBasicAnalysis.get_statistic_result_table(first_col_values, df_count,
                                                                                statistic_result,
                                                                                statistic_tests_options)

        box_plot_title = "Data Chart"
        if len(self.input_df) > 10000:
            # Sampling only 10000 rows
            box_plot_title += " (Sampling 10000 rows)"
            for index, df in enumerate(statistic_chart):
                statistic_chart[index] = df.copy().sample(n=10000, random_state=1, replace=False)

        box_plot_plotly = WidgetBasicAnalysis.get_box_plot_plotly(statistic_chart)

        return CommonWidget.get_accordion_widget([statistic_result_table, box_plot_plotly],
                                                 ['Statistical table', box_plot_title])



# Optional file
# Dynamic metadata: possibly non-deterministic. Any items that are dynamic or determined at install-time,
# as well as extension modules or extensions to setuptools, need to go into setup.py.
# Static metadata should be preferred and dynamic metadata should be used only as an escape hatch when absolutely necessary.

import setuptools

setuptools.setup()

"""
import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="jupyter-analysis-extension",
    version="0.0.1",
    author="Kyoungjun Park",
    author_email="parkkjun525@gmail.com",
    description="A Jupyter Extension for Data Analysis",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/KyoungjunPark/ipywidget_statistics",
    project_urls={
        "Bug Tracker": "https://github.com/KyoungjunPark/ipywidget_statistics/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GPL-3.0",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.6",
)
"""


import pandas as pd
from tkinter import Tk, filedialog

from jupyter_analysis_extension.statistics_widget.common_widget import CommonWidget
from jupyter_analysis_extension.utils.widget_maker import WidgetMaker
import os

try:
    from mlplatform_lib.mllab import MllabApi
except ModuleNotFoundError:
    pass


class FileSelector:
    """ Custom FileSelector for pd.readCSV support """

    def init(self, mode=None, data=None, set_main_tab=None, logger=None):
        # data
        self.df = None

        # widget
        self.file_selector_widget = None
        self.uploaded_progress_widget = None

        # logger
        self.logger = logger

        if mode == "HD":
            hyper = MllabApi()
            hyper_do_list = {"Select DO": "Select DO"}
            hyper_do_list.update(hyper.get_do_list())

            dropdown = CommonWidget.get_dropdown(options=hyper_do_list, default_option="Select DO")
            dropdown.observe(self.on_hyper_do_click, names='value')
            self.file_selector_widget = CommonWidget.get_widget_with_label("HyperData", dropdown,
                                                                           margin="0px 0px 0px 0px")

        elif data is not None:
            # 210817 KJPARK
            if type(data) is pd.DataFrame:
                # Case: Dataframe
                self.df = data
            else:
                # Case: Path
                if os.path.exists(data):
                    self.df = pd.read_csv(data)
                else:
                    self.logger.error("File does not exist on path: %s", data)
                    exit(1)
            set_main_tab(self.df)
        else:
            self.file_selector_widget = WidgetMaker.get_btn(description="File select")
            self.file_selector_widget.on_click(self.on_file_selector_click)

        self.uploaded_progress_widget = WidgetMaker.get_html(
            value=" <div></div>",
            style=dict(margin="0px 0px 0px 10px"),
        )
        # callback
        self.set_main_tab = set_main_tab

    def on_file_selector_click(self, button):
        """ get file path by tkinter library & readCSV logic """
        self.set_file_uploading()
        root = Tk()
        root.withdraw()  # Hide the main window.
        root.call('wm', 'attributes', '.', '-topmost', True)  # Raise the root to the top of all windows.
        button.files = filedialog.askopenfilename(title="Select file", filetypes=(
        ("CSV Files", "*.csv"),))  # List of selected files will be set button's file attribute.

        if len(list(button.files)) == 0:
            return self.set_file_uploading_finish()

        self.df = pd.read_csv(str(button.files), encoding='utf-8')
        self.set_main_tab(self.df)
        self.set_file_uploading_finish()

    def on_hyper_do_click(self, value):
        """ get DataObject by Hyperdata python library & Change dataFrame """
        if value['new'] == "Select DO":
            return
        print(int(value['new']))
        hyper = MllabApi()
        self.df = hyper.get_do(value['new'])
        self.set_main_tab(self.df)
        self.set_file_uploading_finish()

    def set_file_uploading(self):
        self.file_selector_widget.description = "Uploading.."
        self.uploaded_progress_widget.add_class("load")

    def set_file_uploading_finish(self):
        self.file_selector_widget.description = "File select"
        self.uploaded_progress_widget.remove_class("load")

    def get_file_selector_container(self):
        file_selector_container = []
        if self.file_selector_widget is not None:
            file_selector_container.append(self.file_selector_widget)

        if self.uploaded_progress_widget is not None:
            file_selector_container.append(self.uploaded_progress_widget)

        return WidgetMaker.get_vertical_box(
            box_items=file_selector_container,
            style=dict(display="flex", width="100%", align_items="center")
        )


#!/usr/bin/env python
# coding: utf-8

# Copyright (c) TmaxEnterprise.
# Distributed under the terms of the Modified BSD License.
import ipywidgets as widgets
import plotly.graph_objs as go


class WidgetLogisticRegression:
    @staticmethod
    def get_statistic_result_table(table):

        res_table = widgets.HTML(table.as_html())
        statistic_result_table = widgets.VBox([res_table])
        statistic_result_table.add_class('styledTable')

        return statistic_result_table

    @staticmethod
    def get_all_scatter_plot(data, columns):
        dim_list = []
        for column in columns:
            dim_dict = {'label': column, 'values': data[column]}
            dim_list.append(dim_dict)
        fig = go.FigureWidget()
        # fig.add_trace(go.Splom(dimensions=[dict(label='Period', values=data['Period']), dict(label='Data_value', values=data['Data_value'])]))
        fig.add_trace(go.Splom(dimensions=dim_list))
        fig.update_layout(
            title='Scatter Plot',
            title_x=0.5,
            title_font_size=30,
            dragmode='select',
            width=600,
            height=600,
            hovermode='closest',
        )

        return fig

    @staticmethod
    def get_heatmap_plot(data, columns):
        fig = go.FigureWidget()
        fig.add_trace(go.Heatmap(x=columns, y=columns, z=data.corr()))
        fig.update_layout(
            title='Heatmap of Variables',
            title_x=0.5,
            title_font_size=30,
            dragmode='select',
            width=600,
            height=600,
            hovermode='closest',
        )
        return fig

    @staticmethod
    def get_eda_plot(data, columns):
        scatter_plot = WidgetLogisticRegression.get_all_scatter_plot(data, columns)
        heatmap_plot = WidgetLogisticRegression.get_heatmap_plot(data, columns)
        return widgets.HBox([scatter_plot, heatmap_plot])



from collections import namedtuple

import numpy as np
import scipy.special as special
from pandas.api.types import is_numeric_dtype
from statsmodels.stats.outliers_influence import variance_inflation_factor


class CommonUtil:
    bool_cond_list = ["0", "1", "x", "X", "y", "Y", "Yes", "No", "yes",
                      "no", "true", "false", "True", "False", "YES", "NO", "TRUE", "FALSE"]
    type_dic = {}
    # distinct_list's elem in bool_cond_list

    @staticmethod
    def get_abstract_type(col):
        # Check whether NaN values exist in columns
        if len(col.dropna().unique())==2:
            abstract_type = '2-Category'
        else:
            if col.isnull().all():
                abstract_type = 'NaN'
            elif col.dtype == np.float or col.dtype == np.int64:
                abstract_type = 'Number'
            elif col.dtype == np.bool:
                abstract_type = '2-Category'
            elif col.dtype == np.object:
                abstract_type = 'Object'
            else:
                abstract_type = 'Object'
        return abstract_type

    @staticmethod
    def get_avail_type_list(column_data):
        uniq = len(column_data.dropna().unique())
        type_list = ['Object', 'N-Category']

        if column_data.isnull().all():
            type_list.append('NaN')
            return type_list

        if is_numeric_dtype(column_data):
            type_list.append('Number')

        if uniq < 3:
            type_list.remove('N-Category')
            if uniq == 2:
                type_list.append('2-Category')
            elif uniq == 1:
                if column_data.isnull().any():
                    type_list.append('2-Category')
                else:
                    type_list.append('Constant')
            else:
                return ['NaN']

        return type_list

    @staticmethod
    def get_column_name_with_type(column_name_list, column_type_list):
        column_with_type = {}
        for column in column_name_list:
            type_name = column_type_list[column]
            if type_name != 'NaN':
                name_with_type = column + " (" + str(type_name) + ")"
                column_with_type[name_with_type] = column
        return column_with_type

    @staticmethod
    def connect_multiples_by_button(control_data, from_title,from_widget, to_title,to_widget, add_btn, delete_btn,
                                    max_count=0):
        def add_btn_clicked(self):
            changed_options = list(to_widget.options)
            changed_options.extend(from_widget.value)
            if max_count > 0:
                changed_options = changed_options[-max_count:]
            to_widget.options = tuple(changed_options)
            control_data[to_title] = changed_options
            from_widget.value = ()
            to_widget.value = ()
        add_btn.on_click(add_btn_clicked)

        def delete_btn_clicked(self):
            changed_options = list(to_widget.options)
            for item in to_widget.value:
                changed_options.remove(item)
            to_widget.options = tuple(changed_options)
            control_data[to_title] = changed_options
            from_widget.value = ()
            to_widget.value = ()
        delete_btn.on_click(delete_btn_clicked)
        return

    @staticmethod
    def update_result_layout(control_panel, result_widget):
        if result_widget is not None:
            target_widget = list(control_panel.children)
            target_widget[2] = result_widget
            control_panel.children = tuple(target_widget)
            return True
        else:
            return False

    @staticmethod
    def check_variables_shapes(control_data, panel_name, target_types, column_type_list, column_value, sub_option):
        """
        Check the requirements of types in the control panel

        Parameters
        ----------
        control_data : dict
        panel_name : str
            Name of the target column.
        target_types
        column_type_list
        column_value
        sub_option : list[dict]
            Contains the limitation of min & max. 0 means no limitation in counts.
        Returns
        -------
        object
        """
        error_msg = None

        if panel_name not in control_data.keys():
            error_msg = 'Please enter the ' + panel_name + '.'
            return error_msg
        min_count = None
        max_count = None
        for sub_op in sub_option:
            if sub_op["name"] == panel_name:
                min_count = (sub_op["min_count"] if "min_count" in sub_op.keys() else None)
                max_count = (sub_op["max_count"] if "max_count" in sub_op.keys() else None)
                break

        if min_count is not None and min_count != 0 and len(control_data[panel_name]) < min_count:
            error_msg = 'Please set the number of ' + panel_name + ' at least ' + str(min_count) + '.'
            return error_msg

        if max_count is not None and max_count != 0 and len(control_data[panel_name]) > max_count:
            error_msg = 'Please set the number of ' + panel_name + ' at most ' + str(max_count) + '.'
            return error_msg

        if panel_name in control_data.keys():
            type_condition = True
            column_list = []
            for item in control_data[panel_name]:
                if str(column_type_list[column_value[item]]) not in target_types:
                    column_list.append(column_value[item])
                    type_condition = False

            if not type_condition:
                if len(column_list) > 1:
                    error_msg = panel_name + ' of [' + ', '.join(column_list) + '] are not ' + str(target_types) + ' type.'
                else:
                    error_msg = panel_name + ' of [' + ', '.join(column_list) + '] is not ' + str(target_types) + ' type.'
        return error_msg

    @staticmethod
    def check_option_checkbox(control_data, option_name, minimum_checked=None):
        error_msg = None
        target_option = control_data["option"][option_name]

        checked_list = []
        if minimum_checked is not None:
            for key in target_option.keys():
                checked_list.append(target_option[key])
            if checked_list.count(True) < minimum_checked:
                error_msg = 'Please check Option-'+option_name+' at least '+str(minimum_checked)
        return error_msg

class StatsUtil:
    @staticmethod
    def welch_anova(*args, var_equal=False):
        # https://svn.r-project.org/R/trunk/src/library/stats/R/oneway.test.R
        # translated from R Welch ANOVA (not assuming equal variance)
        # as cited in https://github.com/scipy/scipy/issues/11122
        F_onewayResult = namedtuple('F_onewayResult', ('statistic', 'pvalue'))
        args = [np.asarray(arg, dtype=float) for arg in args]
        k = len(args)
        ni = np.array([len(arg) for arg in args]).astype(float)
        mi = np.array([np.mean(arg) for arg in args]).astype(float)
        vi = np.array([np.var(arg, ddof=1) for arg in args]).astype(float)
        wi = np.divide(ni, vi, out=np.zeros_like(ni), where=vi!=0)

        tmp = sum((1 - wi / sum(wi)) ** 2 / (ni - 1))
        tmp /= (k ** 2 - 1)

        dfbn = k - 1
        dfwn = 1 / (3 * tmp)

        m = sum(mi * wi) / sum(wi)
        f = sum(wi * (mi - m) ** 2) / ((dfbn) * (1 + 2 * (dfbn - 1) * tmp))
        prob = special.fdtrc(dfbn, dfwn, f)  # equivalent to stats.f.sf
        return F_onewayResult(f, prob)

    @staticmethod
    def vif(X, thresh=5.0):
        variables = list(range(X.shape[1]))
        dropped = True
        while dropped:
            dropped = False
            vif = [variance_inflation_factor(X.iloc[:, variables].values, ix)
                   for ix in range(X.iloc[:, variables].shape[1])]

            maxloc = vif.index(max(vif))
            if max(vif) > thresh:
                print('dropping \'' + X.iloc[:, variables].columns[maxloc] +
                      '\' at index: ' + str(maxloc))
                del variables[maxloc]
                dropped = True

        print('Remaining variables:')
        print(X.columns[variables])
        return X.iloc[:, variables]

    @staticmethod
    def stratified_sample(df, strata, size=10000, seed=None, keep_index=True):
        '''
        It samples data from a pandas dataframe using strata. These functions use
        proportionate stratification:
        n1 = (N1/N) * n
        where:
            - n1 is the sample size of stratum 1
            - N1 is the population size of stratum 1
            - N is the total population size
            - n is the sampling size
        Parameters
        ----------
        :df: pandas dataframe from which data will be sampled.
        :strata: list containing columns that will be used in the stratified sampling.
        :size: sampling size. Fixed to 10000 temporarily.
        :seed: sampling seed
        :keep_index: if True, it keeps a column with the original population index indicator

        Returns
        -------
        A sampled pandas dataframe based in a set of strata.
        '''
        population = len(df)
        tmp = df[strata]
        tmp['size'] = 1
        tmp_grpd = tmp.groupby(strata).count().reset_index()
        tmp_grpd['samp_size'] = round(size / population * tmp_grpd['size']).astype(int)

        # controlling variable to create the dataframe or append to it
        first = True
        for i in range(len(tmp_grpd)):
            # query generator for each iteration
            qry = ''
            for s in range(len(strata)):
                stratum = strata[s]
                value = tmp_grpd.iloc[i][stratum]
                n = tmp_grpd.iloc[i]['samp_size']

                if type(value) == str:
                    value = "'" + str(value) + "'"

                if s != len(strata) - 1:
                    qry = qry + stratum + ' == ' + str(value) + ' & '
                else:
                    qry = qry + stratum + ' == ' + str(value)

            # final dataframe
            if first:
                stratified_df = df.query(qry).sample(n=n, random_state=seed).reset_index(drop=(not keep_index))
                first = False
            else:
                tmp_df = df.query(qry).sample(n=n, random_state=seed).reset_index(drop=(not keep_index))
                stratified_df = stratified_df.append(tmp_df, ignore_index=True)

        return stratified_df




import numpy as np
import ipywidgets as widgets

# Graph
import plotly.graph_objs as go


class WidgetBasicPCA:
    @staticmethod
    def get_pca_variance_chart(variance_ratio_list):
        fig = go.FigureWidget()

        chart_x = []
        for i in range(0, len(variance_ratio_list)):
            chart_x.append("PC"+str(i+1))

        fig.add_trace(
            go.Bar(
                x=chart_x,
                y=variance_ratio_list,
                name='Individual'
            )
        )
        fig.add_trace(
            go.Scatter(
                x=chart_x,
                y=np.cumsum(variance_ratio_list),
                mode='lines+markers',
                name='Cumulative'
            )
        )

        fig.update_layout(
            title='PCA plot',
            title_x=0.5,
            title_font_size=25,
            yaxis_title="Explained variance in percent",
            width=550,
            height=450,
            hovermode='closest',
        )
        return fig

    @staticmethod
    def get_matrix_plot(transform_df, column_list, variance_ratio_list):
        fig = go.FigureWidget()
        dimensions =[]
        for column in column_list:
            dimensions.append({"label" : column, "values" :transform_df[column]})
        total_var = variance_ratio_list.sum() * 100
        fig.add_trace(
            go.Splom(
                dimensions=dimensions,
                diagonal_visible=False,
                marker=dict(
                            showscale=False,
                            line_color='white', line_width=0.5)
            )
        )
        fig.update_layout(
            title=f'Matrix Plot<br>Total Explained Variance: {total_var:.2f}%',
            title_x=0.5,
            title_font_size=25,
            width=450,
            height=450,
        )
        return fig

    @staticmethod
    def get_scree_plot(variance_ratio_list):
        fig = go.FigureWidget()

        chart_x = []
        for i in range(0, len(variance_ratio_list)):
            chart_x.append(str(i+1))

        fig.add_trace(
            go.Scatter(
                x=chart_x,
                y=variance_ratio_list,
                name='Eigenvalue',
                mode='lines+markers',
            )
        )

        fig.add_trace(
            go.Scatter(
                x=chart_x,
                y=[1 for i in range(len(chart_x))],
                mode='lines',
                name="Selection baseline"
            )
        )
        fig.update_layout(
            title='Scree plot',
            title_x=0.5,
            title_font_size=25,
            xaxis_title="Number of Componenets",
            yaxis_title="Eigenvalue",
            width=550,
            height=450,
            hovermode='closest',
        )
        return fig

    @staticmethod
    def get_kmo_standard_table():
        standard_table = widgets.HTML(
            value="""<table style="width: 300px">
                       <thead>
                       <th>Value</th>
                       <th>Status</th>
                       </thead>
                       <tbody>
                           <tr>
                               <td>0.9 > k</td>
                               <td>매우 좋음</td>
                           </tr>
                           <tr>
                               <td>0.8 < k < 0.89</td>
                               <td>꽤 좋음</td>
                           </tr>
                           <tr>
                               <td>0.7 < k < 0.79</td>
                               <td>적당함</td>
                           </tr>
                           <tr>
                               <td>0.6 < k < 0.69</td>
                               <td>평범함</td>
                           </tr>
                           <tr>
                               <td>0.5 < k < 0.59</td>
                               <td>좋지 않음</td>
                           </tr>
                            <tr>
                               <td>k < 0.5</td>
                               <td>사용할 수 없음</td>
                           </tr>
                       </tbody>
                   </table>
                   <div>k = Overall MSA value<div>
                   """,
            disabled=True
        )
        standard_table.add_class('styledTable')

        return standard_table


import ipywidgets as widgets


class WidgetMaker:
    @staticmethod
    def get_vertical_box(box_items=[], style={}):
        return widgets.HBox(
            box_items,
            layout=widgets.Layout(**style),
        )

    @staticmethod
    def get_btn(description="", disabled=False, tooltip="", style={}):
        return widgets.Button(
            description=description,
            disabled=disabled,
            tooltip=tooltip,
            layout=widgets.Layout(**style),
        )

    @staticmethod
    def get_html(value="", style={}):
        return widgets.HTML(
            value=value,
            layout=widgets.Layout(**style),
        )

    @staticmethod
    def get_styled_table(header_list, body_list, footer="", direation="column"):
        table_header_info = ""
        for item in header_list:
            table_header_info += "<th>{0}</th>".format(item)

        table_body_info = ""
        for body_item in body_list:
            table_row_info =""
            for row_item in body_item:
                table_row_info += "<td>{0}</td>".format(row_item)
            table_body_info += "<tr>{0}</tr>".format(table_row_info)

        res_table = widgets.HTML(
            value="""
                <table>
                    <thead>
                    {0}
                    </thead>
                    <tbody>
                    {1}
                    </tbody>
                </table>
                <div>{2}<div>
                """.format(table_header_info,table_body_info,footer),
            disabled=True
        )
        res_table.add_class('styledTable')
        if direation=="row" :
            return widgets.HBox([res_table])
        return widgets.VBox([res_table])

    @staticmethod
    def get_individual_connected_unique_numberField(control_data_option, title, connect_widgets, column_name_type_map, type_list,
                                          df, acc_title):
        individual_connected_unique_numberField = None
        control_data_option[title] = {}

        def on_numberField_changed(value):
            control_data_option[title][value['owner'].description] = value['new']

        def on_multiple_changed(value):
            total_options = []
            column_name = column_name_type_map[connect_widgets[0].options[0]]
            unique_items = df[column_name].unique()
            for item in connect_widgets:
                total_options.extend(list(item.options))

            number_field_widgets = []
            changed_option = {}
            for item in unique_items:
                number_field_widget = widgets.BoundedIntText(
                                        value=1,
                                        min=0,
                                        description=str(item),
                                        disabled=False,
                                        layout=widgets.Layout(width="180px"),

                                    )
                number_field_widgets.append(number_field_widget)
                number_field_widget.observe(on_numberField_changed, names="value")
                individual_connected_unique_numberField.children = number_field_widgets

            control_data_option[title] = changed_option

        for widget in connect_widgets:
            widget.observe(on_multiple_changed, names="options")

        individual_connected_unique_numberField = widgets.VBox([])

        accordion_item = widgets.HBox(
            [individual_connected_unique_numberField],
            layout=widgets.Layout(display="flex", justify_content="flex-start", margin="0", width="calc(100% + 5px)"))
        connected_multiple = WidgetMaker.get_accordion_widget([accordion_item], [acc_title], selected_index=None,
                                                               style=dict(width="100%", display="flex", justify_content='center', margin="0",max_height="500px"))

        return widgets.VBox([connected_multiple], layout=widgets.Layout(margin='0px 0px 0px 10px'))


    @staticmethod
    def get_accordion_widget(widget_list, title_list, selected_index=0,style={}):
        accordion_widget = widgets.Accordion(
            children=widget_list,
            layout=widgets.Layout(**style),
            selected_index=selected_index)
        for i in range(len(accordion_widget.children)):
            accordion_widget.set_title(i, title_list[i])
        return accordion_widget


import ipywidgets as widgets
import logging


class OutputWidgetHandler(logging.Handler):
    """ Custom logging handler sending logs to an output widget """

    def __init__(self, *args, **kwargs):
        super(OutputWidgetHandler, self).__init__(*args, **kwargs)
        layout = {
            'width': '100%',
            'height': '160px',
            'border': '1px solid black'
        }
        self.out = widgets.Output(layout=layout)

    def emit(self, record):
        """ Overload of logging.Handler method """
        formatted_record = self.format(record)
        new_output = {
            'name': 'stdout',
            'output_type': 'stream',
            'text': formatted_record+'\n'
        }
        self.out.outputs = (new_output, ) + self.out.outputs

    def show_logs(self):
        """ Show the logs """
        display(self.out)

    def clear_logs(self):
        """ Clear the current logs """
        self.out.clear_output()
